@@ -4,83 +4,105 @@ use strict;
 use warnings;
 use 5.010;
 
-our $VERSION = '1.07';
+our $VERSION = '1.08';
 
-use Getopt::Long;
 use Net::MPD;
 use Proc::Daemon;
 use Log::Dispatch;
+use AppConfig;
 
 sub new {
   my ($class, @options) = @_;
 
   my $self = bless {
-    action     => undef,
-    after      => 2,
-    before     => 2,
-    calls_path => 'calls',
-    calls_freq => 3600,
-    daemon     => 1,
-    last_call  => 0,
-    mpd        => undef,
-    mpd_conn   => 'localhost',
-    music_path => 'music',
+    last_call     => 0,
+    config_errors => [],
     @options
   }, $class;
 }
 
-sub mpd {
-  my ($self) = @_;
-
-  $self->{mpd};
-}
+sub mpd    { shift->{mpd} }
+sub log    { shift->{log} }
+sub config { shift->{config} }
 
 sub parse_options {
-  my ($self, @options) = @_;
+  my ($self, @args) = @_;
 
-  local @ARGV = @options;
+  $self->{config} = AppConfig->new({
+      ERROR => sub { push @{ $self->{config_errors} }, \@_; },
+      CASE  => 1,
+    },
+    'conf|f=s' => {
+      VALIDATE => sub { -e shift }
+    },
+    'before|b=i'   => { DEFAULT => 2 },
+    'after|a=i'    => { DEFAULT => 2 },
+    'calls-path=s' => { DEFAULT => 'calls' },
+    'calls-freq=i' => { DEFAULT => 3600 },
+    'daemon|D!'    => { DEFAULT => 1 },
+    'mpd=s'        => { DEFAULT => 'localhost' },
+    'music-path=s' => { DEFAULT => 'music' },
+    'syslog|s=s'   => { DEFAULT => '' },
+    'conlog|l=s'   => { DEFAULT => '' },
+    'help|h'       => { ACTION  => \&help, },
+    'version|V'    => { ACTION  => \&version, });
 
-  Getopt::Long::Configure('bundling');
-  Getopt::Long::GetOptions(
-    'D|daemon!'      => \$self->{daemon},
-    'V|version'      => sub { $self->{action} = 'show_version' },
-    'a|after=i'      => \$self->{after},
-    'b|before=i'     => \$self->{before},
-    'calls-path=s'   => \$self->{calls_path},
-    'c|calls-freq=i' => \$self->{calls_freq},
-    'h|help'         => sub { $self->{action} = 'show_help' },
-    'mpd=s'          => \$self->{mpd_conn},
-    'music-path=s'   => \$self->{music_path},
-    's|syslog=s'     => \$self->{syslog},
-    'l|conlog=s'     => \$self->{conlog},
-  );
+  $self->_getopt(@args);    # to get --conf option, if any
+
+  my @configs =
+    $self->config->get('conf') || ('/etc/mpdj.conf', "$ENV{HOME}/.mpdjrc");
+  foreach my $config (@configs) {
+    if (-e $config) {
+      say "Loading config ($config)" if $self->config->get('conlog');
+      $self->config->file($config);
+    } else {
+      say "Config file skipped ($config)" if $self->config->get('conlog');
+    }
+  }
+
+  $self->_getopt(@args);    # to override config file
+}
+
+sub _getopt {
+  my ($self, @args) = @_;
+
+  $self->config->getopt([@args]);    # do not consume @args
+
+  if (@{ $self->{config_errors} }) {
+    foreach my $err (@{ $self->{config_errors} }) {
+      printf STDERR @$err;
+      print STDERR "\n";
+    }
+    $self->help;
+  }
 }
 
 sub connect {
   my ($self) = @_;
 
-  $self->{mpd} = Net::MPD->connect($self->{mpd_conn});
+  $self->{mpd} = Net::MPD->connect($self->config->get('mpd'));
 }
 
 sub execute {
   my ($self) = @_;
 
-  if (my $action = $self->{action}) {
-    $self->$action() and return 1;
-  }
-
-  @SIG{qw( INT TERM HUP )} = sub { $self->safe_exit() };
+  local @SIG{qw( INT TERM HUP )} = sub {
+    $self->log->notice('Exiting');
+    exit 0;
+  };
 
   my @loggers;
-  push @loggers, ([ 'Screen', min_level => $self->{conlog}, newline => 1 ])
-    if $self->{conlog};
-  push @loggers, ([ 'Syslog', min_level => $self->{syslog}, ident => 'mpdj' ])
-    if $self->{syslog};
+  push @loggers,
+    ([ 'Screen', min_level => $self->config->get('conlog'), newline => 1 ])
+    if $self->config->get('conlog');
+  push @loggers,
+    ([ 'Syslog', min_level => $self->config->get('syslog'), ident => 'mpdj' ])
+    if $self->config->get('syslog');
 
   $self->{log} = Log::Dispatch->new(outputs => \@loggers);
 
-  if ($self->{daemon}) {
-    $self->{log}->notice('Forking to background');
+  if ($self->config->get('daemon')) {
+    $self->log->notice('Forking to background');
     Proc::Daemon::Init;
   }
 
@@ -92,7 +114,7 @@ sub execute {
   $self->update_cache;
 
   while (1) {
-    $self->{log}->debug('Waiting');
+    $self->log->debug('Waiting');
     my @changes =
       $self->mpd->idle(qw(database player playlist message options));
     $self->mpd->update_status();
@@ -107,35 +129,34 @@ sub execute {
 sub configure {
   my ($self) = @_;
 
-  $self->{log}->notice('Configuring MPD server');
+  $self->log->notice('Configuring MPD server');
 
   $self->mpd->repeat(0);
   $self->mpd->random(0);
 
-  if ($self->{calls_freq}) {
+  if ($self->config->get('calls-freq')) {
     my $now = time;
-    $self->{last_call} = $now - $now % $self->{calls_freq};
-    $self->{log}->notice("Set last call to $self->{last_call}");
+    $self->{last_call} = $now - $now % $self->config->get('calls-freq');
+    $self->log->notice("Set last call to $self->{last_call}");
   }
 }
 
 sub update_cache {
   my ($self) = @_;
 
-  $self->{log}->notice('Updating music and calls cache...');
+  $self->log->notice('Updating music and calls cache...');
 
   foreach my $category (('music', 'calls')) {
 
     @{ $self->{$category} } = grep { $_->{type} eq 'file' }
-      $self->mpd->list_all($self->{"${category}_path"});
+      $self->mpd->list_all($self->config->get("${category}-path"));
 
     my $total = scalar(@{ $self->{$category} });
     if ($total) {
-      $self->{log}
-        ->notice(sprintf("Total %s available: %d", $category, $total));
+      $self->log->notice(sprintf 'Total %s available: %d', $category, $total);
     } else {
-      $self->{log}
-        ->warning("No $category available.  Path is mpd path not file system.");
+      $self->log->warning(
+        "No $category available.  Path should be mpd path not file system.");
     }
   }
 }
@@ -144,9 +165,9 @@ sub remove_old_songs {
   my ($self) = @_;
 
   my $song = $self->mpd->song || 0;
-  my $count = $song - $self->{before};
+  my $count = $song - $self->config->get('before');
   if ($count > 0) {
-    $self->{log}->info("Deleting $count old songs");
+    $self->log->info("Deleting $count old songs");
     $self->mpd->delete("0:$count");
   }
 }
@@ -155,9 +176,10 @@ sub add_new_songs {
   my ($self) = @_;
 
   my $song = $self->mpd->song || 0;
-  my $count = $self->{after} + $song - $self->mpd->playlist_length + 1;
+  my $count =
+    $self->config->get('after') + $song - $self->mpd->playlist_length + 1;
   if ($count > 0) {
-    $self->{log}->info("Adding $count new songs");
+    $self->log->info("Adding $count new songs");
     $self->add_song for 1 .. $count;
   }
 }
@@ -171,13 +193,13 @@ sub add_song {
 sub add_call {
   my ($self) = @_;
 
-  $self->{log}->info('Injecting call');
+  $self->log->info('Injecting call');
 
   $self->add_random_item_from_category('calls', 'immediate');
 
   my $now = time;
-  $self->{last_call} = $now - $now % $self->{calls_freq};
-  $self->{log}->info('Set last call to ' . $self->{last_call});
+  $self->{last_call} = $now - $now % $self->config->get('calls-freq');
+  $self->log->info('Set last call to ' . $self->{last_call});
 }
 
 sub add_random_item_from_category {
@@ -185,13 +207,13 @@ sub add_random_item_from_category {
 
   my @items = @{ $self->{$category} };
 
-  my $index = int(rand(scalar @items));
+  my $index = int rand scalar @items;
   my $item  = $items[$index];
 
   my $uri  = $item->{uri};
   my $song = $self->mpd->song || 0;
   my $pos  = $next ? $song + 1 : $self->mpd->playlist_length;
-  $self->{log}->info('Adding ' . $uri . ' at position ' . $pos);
+  $self->log->info('Adding ' . $uri . ' at position ' . $pos);
 
   $self->mpd->add_id($uri, $pos);
 }
@@ -199,31 +221,22 @@ sub add_random_item_from_category {
 sub time_for_call {
   my ($self) = @_;
 
-  return unless $self->{calls_freq};
-  return time - $self->{last_call} > $self->{calls_freq};
+  return unless $self->config->get('calls-freq');
+  return time - $self->{last_call} > $self->config->get('calls-freq');
 }
 
-sub show_version {
-  my ($self) = @_;
-
+sub version {
   say "mpdj (App::MPDJ) version $VERSION";
+  exit;
 }
 
-sub safe_exit {
-  my ($self) = @_;
-
-  $self->{log}->log_and_die(level => 'notice', message => 'Ending');
-}
-
-sub show_help {
-  my ($self) = @_;
-
-  print <<HELP;
+sub help {
+  print <<'HELP';
 Usage: mpdj [options]
 
 Options:
-  --mpd             MPD connection string (password\@host:port)
-  -s --syslog       Turns on syslog output (debug, info, notice, warn[ing], error, etc)
+  --mpd             MPD connection string (password@host:port)
+  -s,--syslog       Turns on syslog output (debug, info, notice, warn[ing], error, etc)
   -l,--conlog       Turns on console output (same choices as --syslog)
   --no-daemon       Turn off daemonizing
   -b,--before       Number of songs to keep in playlist before current song
@@ -231,9 +244,12 @@ Options:
   -c,--calls-freq   Frequency to inject call signs in seconds
   --calls-path      Path to call sign files
   --music-path      Path to music files
+  -f,--conf         Config file to use
   -V,--version      Show version information and exit
   -h,--help         Show this help and exit
 HELP
+
+  exit;
 }
 
 sub database_changed {
@@ -270,7 +286,7 @@ sub message_changed {
 sub options_changed {
   my $self = shift;
 
-  $self->{log}->notice('Resetting configuration');
+  $self->log->notice('Resetting configuration');
 
   $self->mpd->repeat(0);
   $self->mpd->random(0);
@@ -281,10 +297,11 @@ sub handle_message_mpdj {
 
   my ($option, $value) = split /\s+/, $message, 2;
 
-  if ($option =~ /^(?:before|after|calls_freq)$/) {
+  if ($option eq 'before' or $option eq 'after' or $option eq 'calls-freq') {
     return unless $value =~ /^\d+$/;
-    $self->{log}->info('Setting ' . $option . ' to ' . $value);
-    $self->{$option} = $value;
+    $self->log->info(sprintf 'Setting %s to %s (was %s)',
+      $option, $value, $self->config->get($option));
+    $self->config->set($option, $value);
     $self->player_changed();
   }
 }
@@ -303,7 +320,7 @@ App::MPDJ - MPD DJ.
 
   > mpdj
   > mpdj --before 2 --after 6
-  > mpdj --no-daemon --verbose
+  > mpdj --no-daemon --conlog info
 
 =head1 DESCRIPTION
 
@@ -358,6 +375,10 @@ Path to call sign files.  The default is 'calls'.
 
 Path to music files.  The default is 'music'.
 
+=item -f --conf
+
+Config file to use
+
 =item -V, --version
 
 Show the current version of the script installed and exit.
@@ -368,6 +389,14 @@ Show this help and exit.
 
 =back
 
+=head1 CONFIGURATION FILES
+
+The configuration file is formatted as an INI file.  See L<AppConfig> for
+details.  If no configuration file is given, the file C</etc/mpdj.conf> will be
+read (if it exists) followed by the file C<~/.mpdjrc> (if it exists).  The
+values in the latter file will override anything in the first file.  Command
+line parameters will override anything given in any configuration file.
+
 =head1 AUTHOR
 
 Alan Berndt E<lt>alan@eatabrick.orgE<gt>
